{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzEqcon;

{NAME                            EQUIDISTANT CONIC
PURPOSE:	Transforms input longitude and latitude to Easting and Northing
  for the Equidistant Conic projection.  The longitude and
  latitude must be in radians.  The Easting and Northing values
  will be returned in meters.
PROGRAMMER              DATE
----------              ----
T. Mittan		Mar, 1993
ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
#include "gctpc/cproj.h"
#include "gctpc/report.h"}

interface
uses Math;

//* Variables common to all subroutines in this code file
var
  r_major, //* major axis
    r_minor, //* minor axis
    lon_center, //* Center longitude (projection center)
    lat_origin, //* center latitude
    e0, e1, e2, e3, //* eccentricity constants
    e, es, esp, //* eccentricity constants
    ml0, //* small value m
    false_northing, //* y offset in meters
    false_easting, //* x offset in meters
    ns,
    g,
    rh: double;



function
  eqconforint(r_maj, r_min, lat1, lat2,
  center_lon, center_lat, false_east,
  false_north: double; mode: Integer): Integer;

function
  eqconfor(lon, lat: double; var x, y: double): Integer;

function
  eqconinvint(r_maj, r_min, lat1, lat2,
  center_lon, center_lat, false_east,
  false_north: double; mode: Integer): Integer;

function
  eqconinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;


function
  eqconforint(r_maj, r_min, lat1, lat2,
  center_lon, center_lat, false_east,
  false_north: double; mode: Integer): Integer;
{ Initialize the Equidistant Conic projection
long eqconforint(
double r_maj,/* major axis
double r_min,/* minor axis
double lat1,/* latitude of standard parallel
 double lat2, /* latitude of standard parallel
double center_lon,/* center longitude
 double center_lat,/* center latitude
 double false_east, /* x offset in meters
 double false_north,/* y offset in meters
long int mode) / which format is present A B}
{double e0fn(double x),e1fn(double x),
e2fn(double x),e3fn(double x),
mlfn(double e0, double e1, double e2, double e3, double phi);
 /* functions	*/
double msfnz(double eccent, double sinphi, double cosphi);}
var
  temp, //* temporary variable
    sinphi, cosphi, //* sin and cos values
    ms1, ms2,
    ml1, ml2: double;
begin
// Place parameters in static storage for common use
  r_major := r_maj;
  r_minor := r_min;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
  e0 := e0fn(es);
  e1 := e1fn(es);
  e2 := e2fn(es);
  e3 := e3fn(es);
  sinphi := Sin(lat1);
  cosphi := Cos(lat1);
  {sincos(lat1, sinphi, cosphi);}
  ms1 := msfnz(e, sinphi, cosphi);
  ml1 := mlfn(e0, e1, e2, e3, lat1);

//* format B---------*/
  if (mode <> 0) then
  begin
    if (abs(lat1 + lat2) < EPSLN) then
    begin
      {PError('Standard Parallels on opposite sides of equator',
        'eqcon_for');}
      DoProjError(81,81,0,0,0);
      result := (81);
      exit;
    end;
    sinphi := Sin(lat2);
    cosphi := Cos(lat2);
    {sincos(lat2, sinphi, cosphi);}
    ms2 := msfnz(e, sinphi, cosphi);
    ml2 := mlfn(e0, e1, e2, e3, lat2);
    if (abs(lat1 - lat2) >= EPSLN) then
      ns := (ms1 - ms2) / (ml2 - ml1)
    else ns := sinphi;
  end
  else
    ns := sinphi;
  g := ml1 + ms1 / ns;
  ml0 := mlfn(e0, e1, e2, e3, center_lat);
  rh := r_major * (g - ml0);


//* Report parameters to the user
  if (mode <> 0) then
  begin
    ptitle('EQUIDISTANT CONIC');
    radius2(r_major, r_minor);
    stanparl(lat1, lat2);
    cenlonmer(lon_center);
    origin(center_lat);
    offsetp(false_easting, false_northing);
  end
  else
  begin
    ptitle('EQUIDISTANT CONIC');
    radius2(r_major, r_minor);
    stparl1(lat1);
    cenlonmer(lon_center);
    origin(center_lat);
    offsetp(false_easting, false_northing);
  end;
  result := (OK);
end;

function
  eqconfor(lon, lat: double; var x, y: double): Integer;
{Equidistant Conic forward equations--mapping lat,long to x,y
long eqconfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude 		*/
              /* (I) Latitude 		*/
             /* (O) X projection coordinate 	*/
             /* (O) Y projection coordinate 	*/}
var
  ml,
{double mlfn(double e0, double e1, double e2, double e3, double phi);
double adjust_lon(double x);}
  theta,
    rh1: double;
begin
//* Forward equations  -----------------*/
  ml := mlfn(e0, e1, e2, e3, lat);
  rh1 := r_major * (g - ml);
  theta := ns * AdjustLon(lon - lon_center);
  x := false_easting + rh1 * sin(theta);
  y := false_northing + rh - rh1 * cos(theta);
  result := (OK);
end;



function
  eqconinvint(r_maj, r_min, lat1, lat2,
  center_lon, center_lat, false_east,
  false_north: double; mode: Integer): Integer;
{Initialize the Equidistant Conic projection
  ------------------------------------------*/
long eqconinvint(double r_maj, double r_min, double lat1, double lat2, double center_lon, double center_lat, double false_east, double false_north, long int mode)
/* major axis
/* minor axis
/* latitude of standard parallel
/* latitude of standard parallel
/* center longitude
/* center latitude
/* x offset in meters
/* y offset in meters
/* which format is present A B}
{double e0fn(double x),e1fn(double x),e2fn(double x),e3fn(double x),mlfn(double e0, double e1, double e2, double e3, double phi); 	/* functions	*/
double msfnz(double eccent, double sinphi, double cosphi);}
var
  temp, //* temporary variable
    sinphi, cosphi, //* sin and cos values
    ms1, ms2,
    ml1, ml2: double;
begin
//* Place parameters in static storage for common use
  r_major := r_maj;
  r_minor := r_min;
  lon_center := center_lon;
  lat_origin := center_lat;
  false_northing := false_north;
  false_easting := false_east;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
  e0 := e0fn(es);
  e1 := e1fn(es);
  e2 := e2fn(es);
  e3 := e3fn(es);
    sinphi := Sin(lat1);
    cosphi := Cos(lat1);
  {sincos(lat1, sinphi, cosphi);}
  ms1 := msfnz(e, sinphi, cosphi);
  ml1 := mlfn(e0, e1, e2, e3, lat1);

//* format B---------*/
  if (mode <> 0) then
  begin
    if (abs(lat1 + lat2) < EPSLN) then
    begin
      {PError('Standard Parallels on opposite sides of equator',
        'eqcon-for');}
      DoProjError(81,81,0,0,0);
      result := (81);
      exit;
    end;
    sinphi := Sin(lat2);
    cosphi := Cos(lat2);
    {sincos(lat2, sinphi, cosphi);}
    ms2 := msfnz(e, sinphi, cosphi);
    ml2 := mlfn(e0, e1, e2, e3, lat2);
    if (abs(lat1 - lat2) >= EPSLN) then
      ns := (ms1 - ms2) / (ml2 - ml1)
    else
      ns := sinphi;
  end
  else
    ns := sinphi;
  g := ml1 + ms1 / ns;
  ml0 := mlfn(e0, e1, e2, e3, center_lat);
  rh := r_major * (g - ml0);


//* Report parameters to the user
  if (mode <> 0) then
  begin
    ptitle('EQUIDISTANT CONIC');
    radius2(r_major, r_minor);
    stanparl(lat1, lat2);
    cenlonmer(lon_center);
    origin(center_lat);
    offsetp(false_easting, false_northing);
  end
  else
  begin
    ptitle('EQUIDISTANT CONIC');
    radius2(r_major, r_minor);
    stparl1(lat1);
    cenlonmer(lon_center);
    origin(center_lat);
    offsetp(false_easting, false_northing);
  end;

  result := (OK);
end;


function
  eqconinv(x, y: double; var lon, lat: double): Integer;
{ Equidistant Conic inverse equations--mapping x,y to lat/long
  -----------------------------------------------------------*/
long eqconinv(double x, double y, double *lon, double *lat)
            /* (O) X projection coordinate 	*/
            /* (O) Y projection coordinate 	*/
               /* (I) Longitude 		*/
               /* (I) Latitude 		*/}
{double adjust_lon(double x);
double phi3z(double ml, double e0,
double e1, double e2,
double e3, long int *flag);}
var
  rh1,
    ml,
    con,
    theta: double;
  flag: Integer;
begin
//* Inverse equations  -----------------*/
  flag := 0;
  x := x - false_easting;
  y := rh - y + false_northing;
  if (ns >= 0) then
  begin
    rh1 := sqrt(x * x + y * y);
    con := 1.0;
  end
  else
  begin
    rh1 := -sqrt(x * x + y * y);
    con := -1.0;
  end;
  theta := 0.0;
  if (rh1 <> 0.0) then theta := arctan(con * x / con * y);
  ml := g - rh1 / r_major;
  lat := phi3z(ml, e0, e1, e2, e3, flag);
  lon := AdjustLon(lon_center + theta / ns);

  if (flag <> 0) then result := (flag)
  else
    result := (OK);
end;

end.
