{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzGnom;
{********************************************************************
NAME                            GNOMONIC

PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Gnomonic projection.  The longitude
  and latitude must be in radians.  The Easting and
  Northing values will be returned in meters.

PROGRAMMER              DATE
----------              ----
T. Mittan		Mar, 1993

This function was adapted from the Gnomonic projection code (FORTRAN)
in the General Cartographic Transformation Package software which is
available from the U.S. Geological Survey National Mapping Division.

ALGORITHM REFERENCES

1.  "New Equal-Area Map Projections for Noncircular Regions", John P. Snyder,
    The American Cartographer, Vol 15, No. 4, October 1988, pp. 341-355.

2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.

3.  "Software Documentation for GCTP General Cartographic Transformation
    Package", U.S. Geological Survey National Mapping Division, May 1982.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"  }

interface
uses Math;

//* Variables common to all subroutines in this code file
var
  lon_center, //* Center longitude (projection center)
    lat_center, //* Center latitude (projection center)
    R, //* Radius of the earth (sphere)
    sin_p13, //* Sine of the center latitude
    cos_p13, //* Cosine of the center latitude
    false_easting, //* x offset in meters
    false_northing: Double; //* y offset in meters

function
  gnomforint(r, center_long, center_lat,
  false_east, false_north: double): Integer;

function
  gnomfor(lon, lat: double; var x, y: double): Integer;

function
  gnominvint(r, center_long, center_lat, false_east,
  false_north: double): Integer;

function
  gnominv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

function
  gnomforint(r, center_long, center_lat,
  false_east, false_north: double): Integer;
{Initialize the Gnomonic projection
long gnomforint(double r, double center_long,
double center_lat, double false_east, double false_north)
             /* (I) Radius of the earth (sphere)
                     /* (I) Center longitude
                    /* (I) Center latitude
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use
  R := r;
  lon_center := center_long;
  lat_center := center_lat;
  false_easting := false_east;
  false_northing := false_north;
    sin_p13 := Sin(center_lat);
    cos_p13 := Cos(center_lat);
  {sincos(center_lat, sin_p13, cos_p13);}

//* Report parameters to the user
  ptitle('GNOMONIC');
  radius(r);
  cenlon(center_long);
  cenlat(center_lat);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  gnomfor(lon, lat: double; var x, y: double): Integer;
{Gnomonic forward equations--mapping lat,long to x,y
long gnomfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude
              /* (I) Latitude
             /* (O) X projection coordinate
             /* (O) Y projection coordinate}
{double adjust_lon(double x);}
var
  dlon,
    sinphi, cosphi,
    coslon,
    g,
    ksp: double;
begin
// Forward equations  -----------------*/
  dlon := AdjustLon(lon - lon_center);
    sinphi := Sin(lat);
    cosphi := Cos(lat);
  {sincos(lat, sinphi, cosphi);}
  coslon := cos(dlon);
  g := sin_p13 * sinphi + cos_p13 * cosphi * coslon;
  if (g <= 0.0) then
  begin
    {PError('Point projects into infinity', 'gnomfor-conv');}
    DoProjError(133,133,0,0,0);
    result := (133);
    exit;
  end;
  ksp := 1.0 / g;
  x := false_easting + R * ksp * cosphi * sin(dlon);
  y := false_northing + R * ksp * (cos_p13 * sinphi - sin_p13 * cosphi
    *
    coslon);
  result := (OK);
end;


function
  gnominvint(r, center_long, center_lat, false_east,
  false_north: double): Integer;

{Initialize the Gnomonic projection
long gnominvint(double r, double center_long,
double center_lat, double false_east, double false_north)
             /* (I) Radius of the earth (sphere)
                     /* (I) Center longitude
                    /* (I) Center latitude
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use
  R := r;
  lon_center := center_long;
  lat_center := center_lat;
  false_easting := false_east;
  false_northing := false_north;
    sin_p13 := Sin(center_lat);
    cos_p13 := Cos(center_lat);
  {sincos(center_lat, sin_p13, cos_p13);}

//* Report parameters to the user  -----------------------------*/
  ptitle('GNOMONIC');
  radius(r);
  cenlon(center_long);
  cenlat(center_lat);
  offsetp(false_easting, false_northing);
  result := (OK);
end;

function
  gnominv(x, y: double; var lon, lat: double): Integer;
{Gnomonic inverse equations--mapping x,y to lat/long
long gnominv(double x, double y, double *lon, double *lat)
            /* (O) X projection coordinate */
            /* (O) Y projection coordinate */
               /* (I) Longitude */
               /* (I) Latitude */ }
{double adjust_lon(double x);}
var
  rh,
    z, sinz, cosz,
{double asinz(double con);}
  con: double;
begin
//* Inverse equations  -----------------*/
  x := x - false_easting;
  y := y - false_northing;
  rh := sqrt(x * x + y * y);
  z := arctan(rh / R);
    sinz := Sin(z);
    cosz := Cos(z);
  {sincos(z, sinz, cosz);}
  lon := lon_center;

  if (abs(rh) <= EPSLN) then
  begin
    lat := lat_center;
    result := (OK);
    exit;
  end;
  lat := asinz(cosz * sin_p13 + (y * sinz * cos_p13) / rh);
  con := abs(lat_center) - HALF_PI;
  if (abs(con) <= EPSLN) then
  begin
    if (lat_center >= 0.0) then
    begin
      lon := AdjustLon(lon_center + arctan(x / -y));
      result := (OK);
      exit;
    end
    else
    begin
      lon := AdjustLon(lon_center - arctan(-x / y));
      result := (OK);
      exit;
    end;
  end;
  con := cosz - sin_p13 * sin(lat);

  if ((abs(con) < EPSLN) and (abs(x) < EPSLN)) then begin
    result := (OK);
    exit;
  end;
  lon := AdjustLon(lon_center + arctan((x * sinz * cos_p13) / (con *
    rh)));


  result := (OK);
end;


end.
