{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzGvnsp;

{********************************************************************
NAME                 GENERAL VERTICAL NEAR-SIDE PERSPECTIVE

PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the General Vertical Near-Side Perspective
  projection.  The longitude and latitude must be in
  radians.  The Easting and Northing values will be
  returned in meters.

PROGRAMMER              DATE
----------              ----
T. Mittan		Mar, 1993

This function was adapted from the General Vertical Near-Side Perspective
projection code (FORTRAN) in the General Cartographic Transformation Package
software which is available from the U.S. Geological Survey National Mapping
Division.

ALGORITHM REFERENCES

1.  "New Equal-Area Map Projections for Noncircular Regions", John P. Snyder,
    The American Cartographer, Vol 15, No. 4, October 1988, pp. 341-355.

2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.

3.  "Software Documentation for GCTP General Cartographic Transformation
    Package", U.S. Geological Survey National Mapping Division, May 1982.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}

interface
uses Math;

//* Variables common to all subroutines in this code file
var
  lon_center, //* Center longitude (projection center)
    lat_center, //* Center latitude (projection center)
    RadiusES, //* Radius of the earth (sphere)
    pHeightAboveSphere, //* Height above sphere
    sin_p15, //* Sine of the center latitude
    cos_p15, //* Cosine of the center latitude
    false_easting, //* x offset in meters
    false_northing: Double; //* y offset in meters


function
  gvnspforint(r, h, center_long, center_lat,
  false_east, false_north: double): Integer;

function
  gvnspfor(lon, lat: double; var x, y: double): Integer;

function
  gvnspinvint(r, h, center_long, center_lat,
  false_east, false_north: double): Integer;

function
  gvnspinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

function
  gvnspforint(r, h, center_long, center_lat,
  false_east, false_north: double): Integer;
{Initialize the General Vertical Near-Side Perspective projection
  ---------------------------------------------------------------*/
long gvnspforint(double r, double h, double center_long, double center_lat, double false_east, double false_north)
             /* (I) Radius of the earth (sphere)
            /* height above sphere
                     /* (I) Center longitude
                    /* (I) Center latitude
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use
  RadiusES := r;
  pHeightAboveSphere := 1.0 + h / RadiusES;
  lon_center := center_long;
  lat_center := center_lat;
  false_easting := false_east;
  false_northing := false_north;
    sin_p15 := Sin(center_lat);
    cos_p15 := Cos(center_lat);
  {sincos(center_lat, sin_p15, cos_p15);}

// Report parameters to the user
  ptitle('GENERAL VERTICAL NEAR-SIDE PERSPECTIVE');
  radius(RadiusES);
  genrpt(h, 'Height of Point Above Surface of Sphere:   ');
  cenlon(center_long);
  cenlat(center_lat);
  offsetp(false_easting, false_northing);
  Result := (OK);
end;

function
  gvnspfor(lon, lat: double; var x, y: double): Integer;
{General Vertical Near-Side Perspective forward equations--mapping
   lat,long to x,y
long gvnspfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude */
              /* (I) Latitude */
             /* (O) X projection coordinate */
             /* (O) Y projection coordinate */ }
{double adjust_lon(double x);}
var
  dlon,
    sinphi, cosphi,
    coslon,
    g,
    ksp: double;
begin
//* Forward equations  -----------------*/
  dlon := AdjustLon(lon - lon_center);
    sinphi := Sin(lat);
    cosphi := Cos(lat);
  {sincos(lat, sinphi, cosphi);}
  coslon := cos(dlon);
  g := sin_p15 * sinphi + cos_p15 * cosphi * coslon;
  if (g < (1.0 / pHeightAboveSphere)) then
  begin
    {PError('Point cannot be projected', 'gvnsp-for');}
    DoProjError(153,153,0,0,0);
    Result := (153);
    exit;
  end;
  ksp := (pHeightAboveSphere - 1.0) / (pHeightAboveSphere - g);
  x := false_easting + RadiusES * ksp * cosphi * sin(dlon);
  y := false_northing + RadiusES * ksp * (cos_p15 * sinphi - sin_p15
    * cosphi * coslon);

  Result := (OK);
end;



function
  gvnspinvint(r, h, center_long, center_lat,
  false_east, false_north: double): Integer;
{Initialize the General Vertical Near-Side Perspective projection
  ---------------------------------------------------------------*/
long gvnspinvint(double r, double h,
double center_long, double center_lat,
double false_east, double false_north)
             /* (I) Radius of the earth (sphere)
            /* height above sphere
                     /* (I) Center longitude
                    /* (I) Center latitude
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use
  RadiusES := r;
  pHeightAboveSphere := 1.0 + h / RadiusES;
  lon_center := center_long;
  lat_center := center_lat;
  false_easting := false_east;
  false_northing := false_north;
    sin_p15 := Sin(center_lat);
    cos_p15 := Cos(center_lat);
  {sincos(center_lat, sin_p15, cos_p15);}

//* Report parameters to the user
  ptitle('GENERAL VERTICAL NEAR-SIDE PERSPECTIVE');
  radius(r);
  genrpt(h, 'Height of Point Above Surface of Sphere:    ');
  cenlon(center_long);
  cenlat(center_lat);
  offsetp(false_easting, false_northing);
  Result := (OK);
end;

function
  gvnspinv(x, y: double; var lon, lat: double): Integer;
{General Vertical Near-Side Perspective
inverse equations--mapping   x,y to lat/long
long gvnspinv(double x, double y, double *lon, double *lat)
            /* (O) X projection coordinate */
            /* (O) Y projection coordinate */
               /* (I) Longitude */
               /* (I) Latitude */}
var
{double asinz(double con);
double adjust_lon(double x); }
  rh,
    r,
    con,
    com,
    z, sinz, cosz: double;
begin
//* Inverse equations
  x := x - false_easting;
  y := y - false_northing;
  rh := sqrt(x * x + y * y);

  r := rh / RadiusES;
  con := pHeightAboveSphere - 1.0;
  com := pHeightAboveSphere + 1.0;
  if (r > sqrt(con / com)) then
  begin
    {PError('Input data error', 'gvnsp-for'); }
    DoProjError(155,155,0,0,0);
    Result := (155);
    exit;
  end;
  sinz := (pHeightAboveSphere - sqrt(1.0 - (r * r * com) / con)) /
    (con / r + r / con);
  z := asinz(sinz);
    sinz := Sin(z);
    cosz := Cos(z);
  {sincos(z, sinz, cosz);}
  lon := lon_center;
  if (abs(rh) <= EPSLN) then
  begin
    lat := lat_center;
    Result := (OK);
    exit;
  end;
  lat := asinz(cosz * sin_p15 + (y * sinz * cos_p15) / rh);
  con := abs(lat_center) - HALF_PI;
  if (abs(con) <= EPSLN) then
  begin
    if (lat_center >= 0.0) then
    begin
      lon := AdjustLon(lon_center + arctan(x / -y));
      Result := (OK);
      exit;
    end
    else
    begin
      lon := AdjustLon(lon_center - arctan(-x / y));
      Result := (OK);
      exit;
    end;
  end;
  con := cosz - sin_p15 * sin(lat);
  if ((abs(con) < EPSLN) and (abs(x) < EPSLN)) then
  begin
    Result := (OK);
    exit;
  end;
  lon := AdjustLon(lon_center + arctan((x * sinz * cos_p15) / (con *
    rh)));
  Result := (OK);
end;


end.
