{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzMill;

{******************************************************************
NAME                    MILLER CYLINDRICAL

PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Miller Cylindrical projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.

PROGRAMMER              DATE
----------              ----
T. Mittan		March, 1993

This function was adapted from the Lambert Azimuthal Equal Area projection
code (FORTRAN) in the General Cartographic Transformation Package software
which is available from the U.S. Geological Survey National Mapping Division.

ALGORITHM REFERENCES

1.  "New Equal-Area Map Projections for Noncircular Regions", John P. Snyder,
    The American Cartographer, Vol 15, No. 4, October 1988, pp. 341-355.

2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.

3.  "Software Documentation for GCTP General Cartographic Transformation
    Package", U.S. Geological Survey National Mapping Division, May 1982.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface
uses Math;
//* Variables common to all subroutines in this code file
var
  lon_center, //* Center longitude (projection center) */
    RadiusES, //* Radius of the earth (sphere)	 	*/
    false_easting, //* x offset in meters			*/
    false_northing: Double; //* y offset in meters			*/



function
  millforint(r, center_long, false_east, false_north: double):
    Integer;

function
  millfor(lon, lat: double; var x, y: double): Integer;

function
  millinvint(r, center_long, false_east, false_north: double):
    Integer;

function
  millinv(x, y: double; var lon, lat: double): Integer;

implementation
uses buzGlobal;

function
  millforint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the Miller Cylindrical projection
long
millforint(double r, double center_long, double false_east, double false_north)

             /* (I) Radius of the earth (sphere)
                     /* (I) Center longitude
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use

  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;

//* Report parameters to the user
  ptitle('MILLER CYLINDRICAL');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;

function
  millfor(lon, lat: double; var x, y: double): Integer;
{Miller Cylindrical forward equations--mapping lat,long to x,y
  ------------------------------------------------------------*/
long
millfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude */
              /* (I) Latitude */
             /* (O) X projection coordinate */
             /* (O) Y projection coordinate */}
var dlon: Double;
begin
{double dlon;
double adjust_lon(double x);}

//* Forward equations  -----------------*/
  dlon := AdjustLon(lon - lon_center);
  x := false_easting + RadiusES * dlon;
  y := false_northing + RadiusES * ln(tan((PI / 4.0) + (lat / 2.5)))
    * 1.25;

  result := (OK);
end;


function
  millinvint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the Miller Cylindrical projection
  -------------------------------------------*/
long
millinvint(double r, double center_long, double false_east, double false_north)

             /* (I) Radius of the earth (sphere)
                     /* (I) Center longitude
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use

  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;

//* Report parameters to the user
  ptitle('MILLER CYLINDRICAL');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;

function
  millinv(x, y: double; var lon, lat: double): Integer;
{Miller Cylindrical inverse equations--mapping x,y to lat/long
  ------------------------------------------------------------*/
long millinv(double x, double y, double *lon, double *lat)
            /* (O) X projection coordinate */
            /* (O) Y projection coordinate */
               /* (I) Longitude */
               /* (I) Latitude */}
begin
{double dlon;
double adjust_lon(double x);}

//* Inverse  equations  ------------------*/
  x := x - false_easting;
  y := y - false_northing;

  lon := AdjustLon(lon_center + x / RadiusES);
  lat := 2.5 * (arctan(exp(y / RadiusES / 1.25)) - PI / 4.0);
  result := (OK);
end;


end.
