{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzMolw;

{*********************************************************************
NAME                            MOLLWEIDE

PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the MOllweide projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.

PROGRAMMER              DATE
----------              ----
D. Steinwand, EROS      May, 1991;  Updated Sept, 1992; Updated Feb, 1993
S. Nelson, EDC		Jun, 2993;	Made corrections in precision and
     number of iterations.

ALGORITHM REFERENCES

1.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.

2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface
uses Math;
//* Variables common to all subroutines in this code file
var
  lon_center, //* Center longitude (projection center)
    RadiusES, //* Radius of the earth (sphere)
    false_easting, //* x offset in meters
    false_northing: Double; //* y offset in meters



function
  molwforint(r, center_long, false_east, false_north: double):
    Integer;

function
  molwfor(lon, lat: double; var x, y: double): Integer;

function
  molwinvint(r, center_long, false_east, false_north: double):
    Integer;

function
  molwinv(x, y: double; var lon, lat: double): Integer;

implementation
uses buzGlobal;

function
  molwforint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the Mollweide projection
  ------------------------------------*/
long molwforint(double r, double center_long,
double false_east, double false_north)
             /* (I) Radius of the earth (sphere)
                     /* (I) Center longitude
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use

  false_easting := false_east;
  false_northing := false_north;
  R := r;
  lon_center := center_long;

//* Report parameters to the user

  ptitle('MOLLWEIDE');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  molwfor(lon, lat: double; var x, y: double): Integer;
{Mollweide forward equations--mapping lat,long to x,y
  ----------------------------------------------------*/
long molwfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude */
              /* (I) Latitude */
             /* (O) X projection coordinate */
             /* (O) Y projection coordinate */}
var
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180 */}
  delta_lon, //* Delta longitude (Given longitude - center */
    theta,
    delta_theta,
    con: double;
  i: Integer;
begin
//* Forward equations  -----------------*/
  delta_lon := AdjustLon(lon - lon_center);
  theta := lat;
  con := PI * sin(lat);

//* Iterate using the Newton-Raphson method to find theta

  for i := 0 to 50 do
  begin
    delta_theta := -(theta + sin(theta) - con) / (1.0 + cos(theta));
    theta := theta + delta_theta;
    if (abs(delta_theta) < EPSLN) then break;
    if (i >= 50) then
    begin
      {PError('Iteration failed to converge', 'Mollweide-forward');}
      DoProjError(242,241,0,0,0);
      result := (241);
      exit;
    end;
  end;
  theta := theta / 2.0;

//* If the latitude is 90 deg, force the x coordinate to be "0 + false easting"
//   this is done here because of precision problems with "cos(theta)"
  if (PI / 2 - abs(lat) < EPSLN) then delta_lon := 0;
  x := (0.900316316158 * RadiusES * delta_lon * cos(theta) +
    false_easting);
  y := (1.4142135623731 * RadiusES * sin(theta) + false_northing);
  result := (OK);
end;


function
  molwinvint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the Mollweide projection
  ------------------------------------*/
long molwinvint(double r, double center_long, double false_east, double false_north)
             /* (I) Radius of the earth (sphere) */
                     /* (I) Center longitude */
                    /* x offset in meters
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use

  false_easting := false_east;
  false_northing := false_north;
  R := r;
  lon_center := center_long;

//* Report parameters to the user
  ptitle('MOLLWEIDE');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  molwinv(x, y: double; var lon, lat: double): Integer;
{Mollweide inverse equations--mapping x,y to lat,long
long molwinv(double x, double y, double *lon, double *lat)
           /* (I) X projection coordinate */
           /* (I) Y projection coordinate */
              /* (O) Longitude */
              /* (O) Latitude */}

{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180 */}
var theta, arg: double;
begin
//* Inverse equations  -----------------*/
  x := x - false_easting;
  y := y - false_northing;
  arg := y / (1.4142135623731 * RadiusES);

//* Because of division by zero problems, 'arg' can not be 1.0.  Therefore
//   a number very close to one is used instead.

  if (abs(arg) > 0.999999999999) then arg := 0.999999999999;
  theta := arcsin(arg);
  lon := AdjustLon(lon_center + (x / (0.900316316158 * RadiusES *
    cos(theta))));
  if (lon < (-PI)) then lon := -PI;
  if (lon > PI) then lon := PI;
  arg := (2.0 * theta + sin(2.0 * theta)) / PI;
  if (abs(arg) > 1.0) then arg := 1.0;
  lat := arcsin(arg);
  result := (OK);
end;



end.
