{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzObleq;

{**********************************************************************
NAME                          OBLATED EQUAL-AREA

PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Oblated Equal Area projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.

PROGRAM HISTORY
PROGRAMMER              DATE
----------              ----
D. Steinwand            May, 1991

ALGORITHM REFERENCES

1.  "New Equal-Area Map Projections for Noncircular Regions", John P. Snyder,
    The American Cartographer, Vol 15, No. 4, October 1988, pp. 341-355.

2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.

3.  "Software Documentation for GCTP General Cartographic Transformation
    Package", U.S. Geological Survey National Mapping Division, May 1982.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface
uses Math;
var
  lon_center,
    lat_o,
    theta,
    mOBleq,
    nERD,
    RadiusES,
    sin_lat_o,
    cos_lat_o,
    false_easting,
    false_northing: Double;


function
  obleqforint(r, center_long, center_lat, shape_m,
  shape_n, angle, false_east, false_north: double): Integer;

function
  obleqfor(lon, lat: double; var x, y: double): Integer;

function
  obleqinvint(r, center_long, center_lat, shape_m,
  shape_n, angle, false_east, false_north: double): Integer;

function
  obleqinv(x, y: double; var lon, lat: double): Integer;

implementation
uses buzGlobal;

function
  obleqforint(r, center_long, center_lat, shape_m,
  shape_n, angle, false_east, false_north: double): Integer;
{long
obleqforint(double r, double center_long,
double center_lat, double shape_m,
double shape_n, double angle,
double false_east, double false_north)}
begin
//* Place parameters in static storage for common use

  R := r;
  lon_center := center_long;
  lat_o := center_lat;
  mOBleq := shape_m;
  nERD := shape_n;
  theta := angle;
  false_easting := false_east;
  false_northing := false_north;

//* Report parameters to the user (to device set up prior to this call)

  ptitle('OBLATED EQUAL-AREA');
  radius(R);
  cenlon(lon_center);
  cenlat(lat_o);
  genrpt(mOBleq, 'Parameter m:      ');
  genrpt(nERD, 'Parameter n:      ');
  genrpt(theta, 'Theta:      ');
  offsetp(false_easting, false_northing);
  pblank();

//* Calculate the sine and cosine of the latitude of the center of the map
//   and store in static storage for common use.
    sin_lat_o := Sin(lat_o);
    cos_lat_o := Cos(lat_o);
  {sincos(lat_o, sin_lat_o, cos_lat_o);}
  result := (OK);
end;

function
  obleqfor(lon, lat: double; var x, y: double): Integer;
{long
obleqfor(double lon, double lat, double *x, double *y)
             /* (I) Longitude */
             /* (I) Latitude */
            /* (O) X projection coordinate */
            /* (O) Y projection coordinate */}
var
  delta_lon,
    sin_delta_lon,
    cos_delta_lon,
    sin_lat,
    cos_lat,
    z,
{ sin_z,
 cos_z,}
  Az,
    sin_Az,
    cos_Az,
    temp,
    x_prime,
    y_prime,
    Mbig,
    Nbig {,
 diff_angle,
 sin_diff_angle,
 cos_diff_angle}: Double;
begin
//* Forward equations  -----------------*/
  delta_lon := lon - lon_center;
    sin_lat := Sin(lat);
    cos_lat := Cos(lat);
  {sincos(lat, sin_lat, cos_lat);}
    sin_delta_lon := Sin(delta_lon);
    cos_delta_lon := Cos(delta_lon);
  {sincos(delta_lon, sin_delta_lon, cos_delta_lon);}
  z := arccos(sin_lat_o * sin_lat + cos_lat_o * cos_lat *
    cos_delta_lon);
  Az := arctan((cos_lat * sin_delta_lon) / (cos_lat_o * sin_lat -
    sin_lat_o *
    cos_lat * cos_delta_lon)) + theta;

    sin_Az := Sin(Az);
    cos_Az := Cos(Az);
  {sincos(Az, sin_Az, cos_Az); }
  temp := 2.0 * sin(z / 2.0);
  x_prime := temp * sin_Az;
  y_prime := temp * cos_Az;
  Mbig := arcsin(x_prime / 2.0);
  temp := y_prime / 2.0 * cos(Mbig) / cos(2.0 * Mbig / mOBleq);
  Nbig := arcsin(temp);
  y := nERD * RadiusES * sin(2.0 * Nbig / nERD) + false_easting;
  x := mOBleq * RadiusES * sin(2.0 * Mbig / mOBleq) * cos(Nbig) /
    cos(2.0 * Nbig / nERD) + false_northing;
  result := (OK);
end;


function
  obleqinvint(r, center_long, center_lat, shape_m,
  shape_n, angle, false_east, false_north: double): Integer;
{long
obleqinvint(double r, double center_long,
double center_lat, double shape_m,
double shape_n, double angle,
double false_east, double false_north)}
begin
//* Place parameters in static storage for common use
  RadiusES := r;
  lon_center := center_long;
  lat_o := center_lat;
  mOBleq := shape_m;
  nERD := shape_n;
  theta := angle;
  false_easting := false_east;
  false_northing := false_north;

//* Report parameters to the user (to device set up prior to this call)
  ptitle('OBLATED EQUAL-AREA');
  radius(RadiusES);
  cenlon(lon_center);
  cenlat(lat_o);
  genrpt(mOBleq, 'Parameter m:      ');
  genrpt(nERD, 'Parameter n:      ');
  genrpt(theta, 'Theta:      ');
  offsetp(false_easting, false_northing);
  pblank();

//* Calculate the sine and cosine of the latitude of the center of the map
//   and store in static storage for common use.
    sin_lat_o := Sin(lat_o);
    cos_lat_o := Cos(lat_o);
  {sincos(lat_o, sin_lat_o, cos_lat_o);}
  result := (OK);
end;

function
  obleqinv(x, y: double; var lon, lat: double): Integer;
{long
obleqinv(double x, double y, double *lon, double *lat)

           /* (I) X projection coordinate */
           /* (I) Y projection coordinate */
              /* (O) Longitude */
              /* (O) Latitude */}
var
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180 */}
  z,
    sin_z,
    cos_z,
    Az,
{ sin_Az,
 cos_Az, }
  temp,
    x_prime,
    y_prime,
    Mbig,
    Nbig,
    diff_angle,
    sin_diff_angle,
    cos_diff_angle: Double;
begin
//* Inverse equations  -----------------*/
  x := x - false_easting;
  y := y - false_northing;
  Nbig := (nERD / 2.0) * arcsin(y / (nERD * RadiusES));
  temp := x / (mOBleq * RadiusES) * cos(2.0 * Nbig / nERD) /
    cos(Nbig);
  Mbig := (mOBleq / 2.0) * arcsin(temp);
  x_prime := 2.0 * sin(Mbig);
  y_prime := 2.0 * sin(Nbig) * cos(2.0 * Mbig / mOBleq) / cos(Mbig);
  temp := sqrt(x_prime * x_prime + y_prime * y_prime) / 2.0;
  z := 2.0 * arcsin(temp);
  Az := arctan(x_prime / y_prime);
  diff_angle := Az - theta;
    sin_diff_angle := Sin(diff_angle);
    cos_diff_angle := Cos(diff_angle);
  {sincos(diff_angle, sin_diff_angle, cos_diff_angle);}
    sin_z := Sin(z);
    cos_z := Cos(z);
  {sincos(z, sin_z, cos_z);}
  lat := arcsin(sin_lat_o * cos_z + cos_lat_o * sin_z *
    cos_diff_angle);
  lon := AdjustLon(lon_center + arctan((sin_z * sin_diff_angle) /
    (cos_lat_o *
    cos_z - sin_lat_o * sin_z * cos_diff_angle)));
  result := (OK);
end;

end.
