{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzOmer;

{**********************************************************************
NAME                       OBLIQUE MERCATOR (HOTINE)
PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Oblique Mercator projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.
PROGRAMMER              DATE
----------              ----
T. Mittan		Mar, 1993
ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}

interface
uses Math;
//* Variables common to all subroutines in this code file
var
{static double} azimuth,
 {static double} r_major, //* major axis
 {static double} r_minor, //* minor axis
{static double} scale_factor, //* scale factor
{static double} lon_origin, //* center longitude
{static double} lat_origin, //* center latitude
 {static double} e, es, //* eccentricity constants
{static double} false_northing, //* y offset in meters
{static double} false_easting, //* x offset in meters
{static double} sin_p20, cos_p20, //* sin and cos values
{static double} bl,
{static double} al,
{static double} d,
{static double} el, u,
{static double} singam, cosgam,
{static double} sinaz, cosaz: Double;

function
  omerforint(r_maj, r_min, scale_fact, azimuth,
  lon_orig, lat_orig, false_east,
  false_north, lon1, lat1, lon2,
  lat2: double; mode: Integer): Integer;

function
  omerfor(lon, lat: double; var x, y: double): Integer;

function
  omerinvint(r_maj, r_min, scale_fact, azimuth,
  lon_orig, lat_orig, false_east,
  false_north, lon1, lat1, lon2,
  lat2: double; mode: Integer): Integer;

function
  omerinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

function
  omerforint(r_maj, r_min, scale_fact, azimuth,
  lon_orig, lat_orig, false_east,
  false_north, lon1, lat1, lon2,
  lat2: double; mode: Integer): Integer;
{ Initialize the Oblique Mercator  projection
long omerforint(
double r_maj,/* major axis			*/
double r_min,/* minor axis			*/
double scale_fact,/* scale factor			*/
double azimuth,/* azimuth east of north	*/
double lon_orig,/* longitude of origin		*/
double lat_orig,/* center latitude		*/
double false_east,/* x offset in meters		*/
double false_north,/* y offset in meters		*/
double lon1,/* fist point to define central line	*/
double lat1,/* fist point to define central line	*/
double lon2,/* second point to define central line	*/
double lat2,/* second point to define central line	*/
 long int mode)/* which format type A or B	*/
}
var
  temp, //* temporary variable		*/
{double adjust_lon(double x);}
{double asinz(double con);}
  con, com,
{double tsfnz(double eccent, double phi, double sinphi);
  /* function to compute t	*/}
  ts,
    ts1, ts2,
    h, l,
    j, p, dlon,
    f, g, gama,
    sinphi, cosphi: double;
begin
//* Place parameters in static storage for common use
  r_major := r_maj;
  r_minor := r_min;
  scale_factor := scale_fact;
  lat_origin := lat_orig;
  false_northing := false_north;
  false_easting := false_east;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
    sin_p20 := Sin(lat_origin);
    cos_p20 := Cos(lat_origin);
  {sincos(lat_origin, sin_p20, cos_p20);}
  con := 1.0 - es * sin_p20 * sin_p20;
  com := sqrt(1.0 - es);
  bl := sqrt(1.0 + es * power(cos_p20, 4.0) / (1.0 - es));
  al := r_major * bl * scale_factor * com / con;
  if (abs(lat_origin) < EPSLN) then
  begin
{   ts := 1.0;}
    f := 1.0;
    d := 1.0;
    el := 1.0;
  end
  else
  begin
    ts := tsfnz(e, lat_origin, sin_p20);
    con := sqrt(con);
    d := bl * com / (cos_p20 * con);
    if ((d * d - 1.0) > 0.0) then
    begin
      if (lat_origin >= 0.0) then
        f := d + sqrt(d * d - 1.0)
      else
        f := d - sqrt(d * d - 1.0);
    end
    else f := d;
    el := f * power(ts, bl);
  end;

//* Report parameters to the user that are the same for both formats
  ptitle('OBLIQUE MERCATOR (HOTINE)');
  radius2(r_major, r_minor);
  genrpt(scale_factor, 'Scale Factor at C. Meridian:    ');
  offsetp(false_easting, false_northing);

  if (mode <> 0) then
  begin
    g := 0.5 * (f - 1.0 / f);
    gama := asinz(sin(azimuth) / d);
    lon_origin := lon_orig - asinz(g * tan(gama)) / bl;

   //* Report parameters common to format B
    genrpt(azimuth * R2D, 'Azimuth of Central Line:    ');
    cenlon(lon_origin);
    cenlat(lat_origin);
    con := abs(lat_origin);
    if ((con > EPSLN) and (abs(con - HALF_PI) > EPSLN)) then
    begin
    singam := Sin(gama);
    cosgam := Cos(gama);
      {sincos(gama, singam, cosgam);}
    sinaz := Sin(azimuth);
    cosaz := Cos(azimuth);
      {sincos(azimuth, sinaz, cosaz);}
      if (lat_origin >= 0) then
        u := (al / bl) * arctan(sqrt(d * d - 1.0) / cosaz)
      else
        u := -(al / bl) * arctan(sqrt(d * d - 1.0) / cosaz);
    end
    else
    begin
      {PError('Input data error', 'omer-init'); }
      DoProjError(201,201,0,0,0);
      result := (201);
      Exit;
    end
  end
  else
  begin
    sinphi := sin(lat1);
    ts1 := tsfnz(e, lat1, sinphi);
    sinphi := sin(lat2);
    ts2 := tsfnz(e, lat2, sinphi);
    h := power(ts1, bl);
    l := power(ts2, bl);
    f := el / h;
    g := 0.5 * (f - 1.0 / f);
    j := (el * el - l * h) / (el * el + l * h);
    p := (l - h) / (l + h);
    dlon := lon1 - lon2;
    if (dlon < -PI) then lon2 := lon2 - 2.0 * PI;
    if (dlon > PI) then lon2 := lon2 + 2.0 * PI;
    dlon := lon1 - lon2;
    lon_origin := 0.5 * (lon1 + lon2) - arctan(j * tan(0.5 * bl *
      dlon) / p) / bl;
    dlon := AdjustLon(lon1 - lon_origin);
    gama := arctan(sin(bl * dlon) / g);
    azimuth := asinz(d * sin(gama));

   //* Report parameters common to format A
    genrpt(lon1 * R2D, 'Longitude of First Point:    ');
    genrpt(lat1 * R2D, 'Latitude of First Point:    ');
    genrpt(lon2 * R2D, 'Longitude of Second Point:    ');
    genrpt(lat2 * R2D, 'Latitude of Second Point:    ');

    if (abs(lat1 - lat2) <= EPSLN) then
    begin
      {PError('Input data error', 'omer-init');}
      DoProjError(202,202,0,0,0);
      result := (202);
      Exit;
    end
    else
      con := abs(lat1);
    if ((con <= EPSLN) or (abs(con - HALF_PI) <= EPSLN)) then
    begin
      {PError('Input data error', 'omer-init'); }
      DoProjError(202,202,0,0,0);
      result := (202);
      Exit;
    end
    else
      if (abs(abs(lat_origin) - HALF_PI) <= EPSLN) then
      begin
        {PError('Input data error', 'omer-init');}
        DoProjError(202,202,0,0,0);
        result := (202);
        Exit;
      end;
    singam := Sin(gama);
    cosgam := Cos(gama);
    {sincos(gama, singam, cosgam);}
    sinaz := Sin(azimuth);
    cosaz := Cos(azimuth);
    {sincos(azimuth, sinaz, cosaz);}
    if (lat_origin >= 0) then
      u := (al / bl) * arctan(sqrt(d * d - 1.0) / cosaz)
    else
      u := -(al / bl) * arctan(sqrt(d * d - 1.0) / cosaz);
  end;
  result := (OK);
end;

function
  omerfor(lon, lat: double; var x, y: double): Integer;
{Oblique Mercator forward equations--mapping lat,long to x,y
  ----------------------------------------------------------*/
long omerfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude 		*/
              /* (I) Latitude 		*/
             /* (O) X projection coordinate 	*/
             /* (O) Y projection coordinate 	*/}
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180
double tsfnz(double eccent, double phi, double sinphi);
/* function to compute t}
var
{ theta,	}//* angle
  sin_phi, { cos_phi,} //* sin and cos value
 { b, }//* temporary values
 {c,} t, {tq, } //* temporary values
    con, {n, ml,} //* cone constant, small m
    q, us, vl,
    ul, vs,
    s,
    dlon,
    ts1: double;
begin
//* Forward equations  -----------------*/
  sin_phi := sin(lat);
  dlon := AdjustLon(lon - lon_origin);
  vl := sin(bl * dlon);
  if (abs(abs(lat) - HALF_PI) > EPSLN) then
  begin
    ts1 := tsfnz(e, lat, sin_phi);
    q := el / (power(ts1, bl));
    s := 0.5 * (q - 1.0 / q);
    t := 0.5 * (q + 1.0 / q);
    ul := (s * singam - vl * cosgam) / t;
    con := cos(bl * dlon);
    if (abs(con) < 0.0000001) then
    begin
      us := al * bl * dlon;
    end
    else
    begin
      us := al * arctan((s * cosgam + vl * singam) / con) / bl;
      if (con < 0) then
        us := us + PI * al / bl;
    end
  end
  else
  begin
    if (lat >= 0) then ul := singam
    else ul := -singam;
    us := al * lat / bl;
  end;
  if (abs(abs(ul) - 1.0) <= EPSLN) then
  begin
    {PError('Point projects into infinity', 'omer-for');}
    DoProjError(205,205,0,0,0);
    result := (205);
    Exit;
  end;
  vs := 0.5 * al * ln((1.0 - ul) / (1.0 + ul)) / bl;
  us := us - u;
  x := false_easting + vs * cosaz + us * sinaz;
  y := false_northing + us * cosaz - vs * sinaz;
  result := (OK);
end;



function
  omerinvint(r_maj, r_min, scale_fact, azimuth,
  lon_orig, lat_orig, false_east,
  false_north, lon1, lat1, lon2,
  lat2: double; mode: Integer): Integer;
{ Initialize the Oblique Mercator  projection
long omerinvint(
double r_maj,/* major axis			*/
double r_min,/* minor axis			*/
double scale_fact,/* scale factor			*/
double azimuth,/* azimuth east of north	*/
double lon_orig,/* longitude of origin		*/
double lat_orig,/* center latitude		*/
double false_east,/* x offset in meters		*/
double false_north,/* y offset in meters		*/
double lon1,/* fist point to define central line	*/
double lat1,/* fist point to define central line	*/
double lon2,/* second point to define central line	*/
double lat2,/* second point to define central line	*/
long int mode)/* which format type A or B	*/}
var
  temp, //* temporary variable		*/
{double adjust_lon(double x);}
{double asinz(double con);}
  con, com,
{double tsfnz(double eccent, double phi, double sinphi);
/* function to compute t	*/}
  h, l, ts1, ts2, ts, {?}
    j, p, dlon,
    f, g, gama,
    sinphi, cosphi: double;
begin
//* Place parameters in static storage for common use

  r_major := r_maj;
  r_minor := r_min;
  scale_factor := scale_fact;
  lat_origin := lat_orig;
  false_northing := false_north;
  false_easting := false_east;

  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e := sqrt(es);
    sin_p20 := Sin(lat_origin);
    cos_p20 := Cos(lat_origin);
  {sincos(lat_origin, sin_p20, cos_p20);}
  con := 1.0 - es * sin_p20 * sin_p20;
  com := sqrt(1.0 - es);
  bl := sqrt(1.0 + es * power(cos_p20, 4.0) / (1.0 - es));
  al := r_major * bl * scale_factor * com / con;
  if (abs(lat_origin) < EPSLN) then
  begin
{   ts := 1.0;}
    f := 1.0;
    d := 1.0;
    el := 1.0;
  end
  else
  begin
    ts := tsfnz(e, lat_origin, sin_p20);
    con := sqrt(con);
    d := bl * com / (cos_p20 * con);
    if ((d * d - 1.0) > 0.0) then
    begin
      if (lat_origin >= 0.0) then
        f := d + sqrt(d * d - 1.0)
      else
        f := d - sqrt(d * d - 1.0);
    end
    else f := d;
    el := f * power(ts, bl);
  end;

//* Report parameters to the user that are the same for both formats
  ptitle('OBLIQUE MERCATOR (HOTINE)');
  radius2(r_major, r_minor);
  genrpt(scale_factor, 'Scale Factor at C. Meridian:    ');
  offsetp(false_easting, false_northing);

  if (mode <> 0) then
  begin
    g := 0.5 * (f - 1.0 / f);
    gama := asinz(sin(azimuth) / d);
    lon_origin := lon_orig - asinz(g * tan(gama)) / bl;

   //* Report parameters common to format B
    genrpt(azimuth * R2D, 'Azimuth of Central Line:    ');
    cenlon(lon_origin);
    cenlat(lat_origin);

    con := abs(lat_origin);
    if ((con > EPSLN) and (abs(con - HALF_PI) > EPSLN)) then
    begin
    singam := Sin(gama);
    cosgam := Cos(gama);
      {sincos(gama, singam, cosgam);}
    sinaz := Sin(azimuth);
    cosaz := Cos(azimuth);
      {sincos(azimuth, sinaz, cosaz);}
      if (lat_origin >= 0) then
        u := (al / bl) * arctan(sqrt(d * d - 1.0) / cosaz)
      else
        u := -(al / bl) * arctan(sqrt(d * d - 1.0) / cosaz);
    end
    else
    begin
      {PError('Input data error', 'omer-init');}
      DoProjError(201,201,0,0,0);
      result := (201);
      Exit;
    end;
  end
  else
  begin
    sinphi := sin(lat1);
    ts1 := tsfnz(e, lat1, sinphi);
    sinphi := sin(lat2);
    ts2 := tsfnz(e, lat2, sinphi);
    h := power(ts1, bl);
    l := power(ts2, bl);
    f := el / h;
    g := 0.5 * (f - 1.0 / f);
    j := (el * el - l * h) / (el * el + l * h);
    p := (l - h) / (l + h);
    dlon := lon1 - lon2;
    if (dlon < -PI) then lon2 := lon2 - 2.0 * PI;
    if (dlon > PI) then lon2 := lon2 + 2.0 * PI;
    dlon := lon1 - lon2;
    lon_origin := 0.5 * (lon1 + lon2) - arctan(j * tan(0.5 * bl *
      dlon) / p) / bl;
    dlon := AdjustLon(lon1 - lon_origin);
    gama := arctan(sin(bl * dlon) / g);
    azimuth := asinz(d * sin(gama));

   //* Report parameters common to format A
    genrpt(lon1 * R2D, 'Longitude of First Point:     ');
    genrpt(lat1 * R2D, 'Latitude of First Point:      ');
    genrpt(lon2 * R2D, 'Longitude of Second Point:    ');
    genrpt(lat2 * R2D, 'Latitude of Second Point:     ');

    if (abs(lat1 - lat2) <= EPSLN) then
    begin
      {PError('Input data error', 'omer-init'); }
      DoProjError(202,202,0,0,0);
      result := (202);
      Exit;
    end
    else
      con := abs(lat1);
    if ((con <= EPSLN) or (abs(con - HALF_PI) <= EPSLN)) then
    begin
      {PError('Input data error', 'omer-init');}
      DoProjError(202,202,0,0,0);
      result := (202);
      Exit;
    end
    else
      if (abs(abs(lat_origin) - HALF_PI) <= EPSLN) then
      begin
       { PError('Input data error', 'omer-init'); }
       DoProjError(202,202,0,0,0);
        result := (202);
        Exit;
      end;
    singam := Sin(gama);
    cosgam := Cos(gama);
    {sincos(gama, singam, cosgam);}
    sinaz := Sin(azimuth);
    cosaz := Cos(azimuth);
    {sincos(azimuth, sinaz, cosaz);}
    if (lat_origin >= 0) then
      u := (al / bl) * arctan(sqrt(d * d - 1.0) / cosaz)
    else
      u := -(al / bl) * arctan(sqrt(d * d - 1.0) / cosaz);
  end;
  result := (OK);
end;



function
  omerinv(x, y: double; var lon, lat: double): Integer;
{ Oblique Mercator inverse equations--mapping x,y to lat/long
long omerinv(double x, double y, double *lon, double *lat)
            /* (O) X projection coordinate 	*/
            /* (O) Y projection coordinate 	*/
               /* (I) Longitude 		*/
               /* (I) Latitude 		*/}
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180}
var
{ delta_lon,}//* Delta longitude (Given longitude - center 	*/
  theta, //* angle					*/
{ delta_theta, }//* adjusted longitude				*/
{ sin_phi, cos_phi,}//* sin and cos value				*/
 { b, }//* temporary values				*/
{ c,} t, { tq,} //* temporary values				*/
    con, {n, ml, } //* cone constant, small m			*/
    vs, us, q, s, ts1,
    vl, ul {,bs,}
{double phi2z(double eccent, double ts, long int *flag);}
{ dlon}: double;
  flag: Integer;
begin
  result := (OK);
//* Inverse equations  -----------------*/
  x := x - false_easting;
  y := y - false_northing;
  flag := 0;
  vs := x * cosaz - y * sinaz;
  us := y * cosaz + x * sinaz;
  us := us + u;
  q := exp(-bl * vs / al);
  s := 0.5 * (q - 1.0 / q);
  t := 0.5 * (q + 1.0 / q);
  vl := sin(bl * us / al);
  ul := (vl * cosgam + s * singam) / t;
  if (abs(abs(ul) - 1.0) <= EPSLN) then
  begin
    lon := lon_origin;
    if (ul >= 0.0) then lat := HALF_PI
    else lat := -HALF_PI;
  end
  else
  begin
    con := 1.0 / bl;
    ts1 := power((el / sqrt((1.0 + ul) / (1.0 - ul))), con);
    lat := phi2z(e, ts1, flag);
    if (flag <> 0) then result := (flag);
    con := cos(bl * us / al);
                         {atan2((s * cosgam - vl * singam) , con)}
    theta := lon_origin - arctan((s * cosgam - vl * singam) / con) /
      bl;
    lon := AdjustLon(theta);
  end;
end;



end.
