{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzRob;

{********************************************************************
NAME                            ROBINSON
PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Robinson projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.
PROGRAMMER              DATE
----------              ----
T. Mittan		March, 1993
This function was adapted from the Lambert Azimuthal Equal Area projection
code (FORTRAN) in the General Cartographic Transformation Package software
which is available from the U.S. Geological Survey National Mapping Division.
ALGORITHM REFERENCES
1.  "New Equal-Area Map Projections for Noncircular Regions", John P. Snyder,
    The American Cartographer, Vol 15, No. 4, October 1988, pp. 341-355.
2.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
3.  "Software Documentation for GCTP General Cartographic Transformation
    Package", U.S. Geological Survey National Mapping Division, May 1982.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h" }


interface

uses Math;

function
  robforint(r, center_long, false_east, false_north: double):
    Integer;

function
  robfor(lon, lat: double; var x, y: double): Integer;

function
  robinvint(r, center_long, false_east, false_north: double):
    Integer;

function
  robinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

//* Variables common to all subroutines in this code file
var
  lon_center, //* Center longitude (projection center) */
    R, //* Radius of the earth (sphere)	 	*/
    false_easting, //* x offset in meters			*/
    false_northing: Double; //* y offset in meters			*/
  pr: array[0..21] of double;
  xlr: array[0..21] of double;

function
  robforint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the ROBINSON projection
long robforint(
double r,/* (I) Radius of the earth (sphere)
double center_long,/* (I) Center longitude
double false_east,/* x offset in meters
double false_north)/* y offset in meters}
var i: Integer;
begin
//* Place parameters in static storage for common use
  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;
  pr[1] := -0.062;
  xlr[1] := 0.9986;
  pr[2] := 0.0;
  xlr[2] := 1.0;
  pr[3] := 0.062;
  xlr[3] := 0.9986;
  pr[4] := 0.124;
  xlr[4] := 0.9954;
  pr[5] := 0.186;
  xlr[5] := 0.99;
  pr[6] := 0.248;
  xlr[6] := 0.9822;
  pr[7] := 0.31;
  xlr[7] := 0.973;
  pr[8] := 0.372;
  xlr[8] := 0.96;
  pr[9] := 0.434;
  xlr[9] := 0.9427;
  pr[10] := 0.4958;
  xlr[10] := 0.9216;
  pr[11] := 0.5571;
  xlr[11] := 0.8962;
  pr[12] := 0.6176;
  xlr[12] := 0.8679;
  pr[13] := 0.6769;
  xlr[13] := 0.835;
  pr[14] := 0.7346;
  xlr[14] := 0.7986;
  pr[15] := 0.7903;
  xlr[15] := 0.7597;
  pr[16] := 0.8435;
  xlr[16] := 0.7186;
  pr[17] := 0.8936;
  xlr[17] := 0.6732;
  pr[18] := 0.9394;
  xlr[18] := 0.6213;
  pr[19] := 0.9761;
  xlr[19] := 0.5722;
  pr[20] := 1.0;
  xlr[20] := 0.5322;

  for i := 0 to 20 do xlr[i] := xlr[i] * 0.9858;

//* Report parameters to the user
  ptitle('ROBINSON');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  robfor(lon, lat: double; var x, y: double): Integer;
{Robinson forward equations--mapping lat,long to x,y
long robfor(
double lon,/* (I) Longitude */
double lat,/* (I) Latitude */
double *x,/* (O) X projection coordinate */
double *y)/* (O) Y projection coordinate */}
var
  dlon,
{double adjust_lon(double x);}
  p2: double;
  ip1: Integer;
begin
//* Forward equations  -----------------*/
  dlon := AdjustLon(lon - lon_center);
  p2 := abs(lat / 5.0 / 0.01745329252);
  ip1 := Round(p2 - EPSLN);

//* Stirling's interpolation formula (using 2nd Diff.)
  p2 := p2 - ip1;
  x := R * (xlr[ip1 + 2] + p2 * (xlr[ip1 + 3] - xlr[ip1 + 1]) / 2.0 +
    p2 * p2 * (xlr[ip1 + 3] - 2.0 * xlr[ip1 + 2] + xlr[ip1 + 1]) /
      2.0) *
    dlon + false_easting;

  if (lat >= 0) then
    y := R * (pr[ip1 + 2] + p2 * (pr[ip1 + 3] - pr[ip1 + 1]) / 2.0 +
      p2 * p2 *
      (pr[ip1 + 3] - 2.0 * pr[ip1 + 2] + pr[ip1 + 1]) / 2.0) * PI /
        2.0 +
      false_northing
  else
    y := -R * (pr[ip1 + 2] + p2 * (pr[ip1 + 3] - pr[ip1 + 1]) / 2.0 +
      p2 * p2 *
      (pr[ip1 + 3] - 2.0 * pr[ip1 + 2] + pr[ip1 + 1]) / 2.0) * PI /
        2.0 +
      false_northing;

  result := (OK);
end;



function
  robinvint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the ROBINSON projection
long robinvint(
double r,/* (I) Radius of the earth (sphere)
 double center_long,/* (I) Center longitude
 double false_east,/* x offset in meters
 double false_north)/* y offset in meters}
var i: Integer;
begin
//* Place parameters in static storage for common use
  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;
  pr[1] := -0.062;
  xlr[1] := 0.9986;
  pr[2] := 0.0;
  xlr[2] := 1.0;
  pr[3] := 0.062;
  xlr[3] := 0.9986;
  pr[4] := 0.124;
  xlr[4] := 0.9954;
  pr[5] := 0.186;
  xlr[5] := 0.99;
  pr[6] := 0.248;
  xlr[6] := 0.9822;
  pr[7] := 0.31;
  xlr[7] := 0.973;
  pr[8] := 0.372;
  xlr[8] := 0.96;
  pr[9] := 0.434;
  xlr[9] := 0.9427;
  pr[10] := 0.4958;
  xlr[10] := 0.9216;
  pr[11] := 0.5571;
  xlr[11] := 0.8962;
  pr[12] := 0.6176;
  xlr[12] := 0.8679;
  pr[13] := 0.6769;
  xlr[13] := 0.835;
  pr[14] := 0.7346;
  xlr[14] := 0.7986;
  pr[15] := 0.7903;
  xlr[15] := 0.7597;
  pr[16] := 0.8435;
  xlr[16] := 0.7186;
  pr[17] := 0.8936;
  xlr[17] := 0.6732;
  pr[18] := 0.9394;
  xlr[18] := 0.6213;
  pr[19] := 0.9761;
  xlr[19] := 0.5722;
  pr[20] := 1.0;
  xlr[20] := 0.5322;
  for i := 0 to 20 do xlr[i] := xlr[i] * 0.9858;

//* Report parameters to the user  -----------------------------*/
  ptitle('ROBINSON');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function robinv(x, y: double; var lon, lat: double): Integer;
{Robinson inverse equations--mapping x,y to lat/long
long   robinv(
double x,/* (O) X projection coordinate */
double y,/* (O) Y projection coordinate */
 double *lon,/* (I) Longitude */
double *lat)/* (I) Latitude */}
{double adjust_lon(double x);}
var
  yy,
    p2,
    u, v, t, c,
    phid,
{ temp, }
  y1: double;
  ip1, i: Integer;
begin

//* Inverse equations  -----------------*/
  x := x - false_easting;
  y := y - false_northing;

  yy := 2.0 * y / PI / R;
  phid := yy * 90.0;
  p2 := abs(phid / 5.0);
  ip1 := round(p2 - EPSLN);
  if (ip1 = 0) then ip1 := 1;

{ Stirling's interpolation formula as used in forward transformation is
   reversed for first estimation of LAT. from rectangular coordinates. LAT.
   is then adjusted by iteration until use of forward series provides correct
   value of Y within tolerance.}
{for (i = 0;;)}
  i := 0;
  repeat
    begin
      u := pr[ip1 + 3] - pr[ip1 + 1];
      v := pr[ip1 + 3] - 2.0 * pr[ip1 + 2] + pr[ip1 + 1];
      t := 2.0 * (abs(yy) - pr[ip1 + 2]) / u;
      c := v / u;
      p2 := t * (1.0 - c * t * (1.0 - 2.0 * c * t));

      if ((p2 >= 0.0) or (ip1 = 1)) then
      begin
        if (y >= 0) then
          phid := (p2 + {(double)} ip1) * 5.0
        else
          phid := -(p2 + {(double)} ip1) * 5.0;
        y1 := y + 1; {added to compile}
        while (abs(y1 - y) > 0.00001) do
        begin
          p2 := abs(phid / 5.0);
          ip1 := round(p2 - EPSLN);
          p2 := p2 - ip1;

          if (y >= 0) then
            y1 := R * (pr[ip1 + 2] + p2 * (pr[ip1 + 3] - pr[ip1 + 1])
              / 2.0 + p2
              * p2 * (pr[ip1 + 3] - 2.0 * pr[ip1 + 2] + pr[ip1 + 1])
                / 2.0)
              * PI / 2.0
          else
            y1 := -R * (pr[ip1 + 2] + p2 * (pr[ip1 + 3] - pr[ip1 + 1])
              / 2.0 + p2
              * p2 * (pr[ip1 + 3] - 2.0 * pr[ip1 + 2] + pr[ip1 + 1])
                / 2.0)
              * PI / 2.0;
          phid := phid + (-180.0 * (y1 - y) / PI / R);
        {i++;} inc(i);
          if (i > 75) then
          begin
            {PError('Too many iterations in inverse', 'robinv-conv');}
            DoProjError(234,234,0,0,0);
            result := (234);
            Exit;
          end;
        end;
      { while (abs(y1 - y) > 0.00001);}
        break;
      end
      else
      begin
        ip1 := ip1 - 1;
        if (ip1 < 0) then
        begin
          {PError('Too many iterations in inverse', 'robinv-conv');}
          DoProjError(234,234,0,0,0);
          result := (234);
          Exit;
        end;
      end;
    end;
  until (I >= 76);
  lat := phid * 0.01745329252;
{ calculate  LONG. using final LAT. with transposed forward Stirling's
   interpolation formula.}
  lon := lon_center + x / R / (xlr[ip1 + 2] + p2 * (xlr[ip1 + 3] -
    xlr[ip1 + 1])
    / 2.0 + p2 * p2 * (xlr[ip1 + 3] - 2.0 * xlr[ip1 + 2] +
    xlr[ip1 + 1]) / 2.0);
  lon := AdjustLon(lon);
  result := (OK);
end;



end.
