{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzSin;

{NAME                  		SINUSOIDAL
PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Sinusoidal projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.
PROGRAMMER              DATE
----------              ----
D. Steinwand, EROS      May, 1991
This function was adapted from the Sinusoidal projection code (FORTRAN) in the
General Cartographic Transformation Package software which is available from
the U.S. Geological Survey National Mapping Division.
ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  "Software Documentation for GCTP General Cartographic Transformation
    Package", U.S. Geological Survey National Mapping Division, May 1982.
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface

uses Math;
//* Variables common to all subroutines in this code file
var
  lon_center, //* Center longitude (projection center)
    R, //* Radius of the earth (sphere)
    false_easting, //* x offset in meters
    false_northing: Double; //* y offset in meters


function
  sinforint(r, center_long, false_east, false_north: double):
    Integer;

function
  sinfor(lon, lat: double; var x, y: double): Integer;

function
  sininvint(r, center_long, false_east, false_north: double):
    Integer;

function
  sininv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;


function
  sinforint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the Sinusoidal projection
long
sinforint(double r, double center_long, double false_east, double false_north)
             /* (I) Radius of the earth (sphere) 	*/
                     /* (I) Center longitude 		*/
                    /* x offset in meters			*/
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use

  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;

//* Report parameters to the user

  ptitle('SINUSOIDAL');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  sinfor(lon, lat: double; var x, y: double): Integer;
{Sinusoidal forward equations--mapping lat,long to x,y
long
sinfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude */
              /* (I) Latitude */
             /* (O) X projection coordinate */
             /* (O) Y projection coordinate */}
var
{double adjust_lon(double x);
 /* Function to adjust longitude to -180 - 180 */}
  delta_lon: double;
 //* Delta longitude (Given longitude - center */
begin
//* Forward equations

  delta_lon := AdjustLon(lon - lon_center);
  x := R * delta_lon * cos(lat) + false_easting;
  y := R * lat + false_northing;
  result := (OK);
end;


function
  sininvint(r, center_long, false_east, false_north: double):
    Integer;
{ Initialize the Sinusoidal projection

long
sininvint(double r, double center_long, double false_east, double false_north)
             /* (I) Radius of the earth (sphere) 	*/
                     /* (I) Center longitude 		*/
                    /* x offset in meters			*/
                     /* y offset in meters}
begin
//* Place parameters in static storage for common use
  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;

//* Report parameters to the user

  ptitle('SINUSOIDAL');
  radius(r);
  cenlon(center_long);
  offsetp(false_easting, false_northing);
  result := (OK);
end;


function
  sininv(x, y: double; var lon, lat: double): Integer;
{Sinusoidal inverse equations--mapping x,y to lat,long
long
sininv(double x, double y, double *lon, double *lat)
           /* (I) X projection coordinate */
           /* (I) Y projection coordinate */
              /* (O) Longitude */
              /* (O) Latitude */}
var
{double adjust_lon(double x);
/* Function to adjust longitude to -180 - 180 */}
  temp: double; //* Re-used temporary variable */
begin
//* Inverse equations

  x := x - false_easting;
  y := y - false_northing;
  lat := y / R;
  if (abs(lat) > HALF_PI) then
  begin
   { PError('Input data error', 'sinusoidal-inverse'); }
    DoProjError(164,164,0,0,0);
    result := (164);
    exit;
  end;
  temp := abs(lat) - HALF_PI;
  if (abs(temp) > EPSLN) then
  begin
    temp := lon_center + x / (R * cos(lat));
    lon := AdjustLon(temp);
  end
  else lon := lon_center;
  result := (OK);
end;


end.
