{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzSom;

{*********************************************************************
NAME                          SPACE OBLIQUE MERCATOR (SOM)
PURPOSE:	The first method of Transforming input longitude and
  latitude to Easting and Northing for the SOM projection.
  The longitude and latitude must be in radians.  The
  Easting and Northing values will be returned in meters.
PROGRAM HISTORY
PROGRAMMER              DATE            REASON
----------              ----            ------
D. Steinwand            July, 1992
T. Mittan		Mar, 1993
S. Nelson		Nov, 1993	Changed error message

ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  "Software Documentation for GCTP General Cartographic Transformation
    Package", U.S. Geological Survey National Mapping Division, May 1982.
*******************************************************************************/
#include <stdio.h>
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface

uses Math;
var
  LANDSAT_RATIO: Double = 0.5201613;
  lon_center, a, b, a2, a4, c1, c3, q, t, u, w, xj, p21, sa, ca, es,
  s, start, false_easting, false_northing: Double;

procedure som_series(var fb, fa2, fa4, fc1, fc3, dlam: double);

function
  somforint(r_major, r_minor: double;
  satnum, path: Integer;
  alf_in, lon, false_east, false_north,
  time: double; start1, flag: Integer): Integer;

function
  somfor(lon, lat: double; var y, x: double): Integer;

function
  sominvint(r_major, r_minor: double;
  satnum, path: Integer;
  alf_in, lon, false_east, false_north,
  time: double; start1, flag: Integer): Integer;

function
  sominv(y, x: double; var lon, lat: double): Integer;


implementation
uses buzGlobal;

function
  somforint(r_major, r_minor: double;
  satnum, path: Integer;
  alf_in, lon, false_east, false_north,
  time: double; start1, flag: Integer): Integer;
{long somforint(double r_major, double r_minor, long int satnum,
 long int path, double alf_in, double lon, double false_east,
 double false_north, double time, long int start1, long int flag)
 /* major axis
 /* minor axis
 /* Landsat satellite number (1,2,3,4,5)
 /* Landsat path number
 /* x offset in meters
 /* y offset in meters}

var
  i: Integer;
  alf, e2c, e2s, one_es,
    dlam, fb, fa2, fa4, fc1, fc3, suma2, suma4, sumc1, sumc3, sumb:
      double;
{double AdjustLon(double x);}
begin
//* Place parameters in static storage for common use
  false_easting := false_east;
  false_northing := false_north;
  a := r_major;
  b := r_minor;
  es := 1.0 - SQR(r_minor / r_major);
  if (flag <> 0) then
  begin
    alf := alf_in;
    p21 := time / 1440.0;
    lon_center := lon;
    start := start1;
  end
  else
  begin
    if (satnum < 4) then
    begin
      alf := 99.092 * D2R;
      p21 := 103.2669323 / 1440.0;
      lon_center := (128.87 - (360.0 / 251.0 * path)) * D2R;
    end
    else
    begin
      alf := 98.2 * D2R;
      p21 := 98.8841202 / 1440.0;
      lon_center := (129.30 - (360.0 / 233.0 * path)) * D2R;
{    lon_center = (-129.30557714 - (360.0/233.0 * path)) * D2R;}
    end;
    start := 0.0;
  end;

// Report parameters to the user (to device set up prior to this call)
  ptitle('SPACE OBLIQUE MERCATOR');
  radius2(a, b);
  if (flag = 0) then
  begin
    GenrptLong(path, 'Path Number:    ');
    GenrptLong(satnum, 'Satellite Number:    ');
  end;
  genrpt(alf * R2D, 'Inclination of Orbit:    ');
  genrpt(lon_center * R2D, 'Longitude of Ascending Orbit:    ');
  offsetp(false_easting, false_northing);
  genrpt(LANDSAT_RATIO, 'Landsat Ratio:    ');

  ca := cos(alf);
  if (abs(ca) < 1.E-9) then ca := 1.E-9;
  sa := sin(alf);
  e2c := es * ca * ca;
  e2s := es * sa * sa;
  w := (1.0 - e2c) / (1.0 - es);
  w := w * w - 1.0;
  one_es := 1.0 - es;
  q := e2s / one_es;
  t := (e2s * (2.0 - es)) / (one_es * one_es);
  u := e2c / one_es;
  xj := one_es * one_es * one_es;
  dlam := 0.0;
  som_series(fb, fa2, fa4, fc1, fc3, dlam);
  suma2 := fa2;
  suma4 := fa4;
  sumb := fb;
  sumc1 := fc1;
  sumc3 := fc3;
{for(i=9;i<=81;i+=18)}
  i := 9; repeat
    begin
      dlam := i;
      som_series(fb, fa2, fa4, fc1, fc3, dlam);
      suma2 := suma2 + 4.0 * fa2;
      suma4 := suma4 + 4.0 * fa4;
      sumb := sumb + 4.0 * fb;
      sumc1 := sumc1 + 4.0 * fc1;
      sumc3 := sumc3 + 4.0 * fc3;
      inc(i, 18);
  end until i <= 81;
{for(i=18; i<=72; i+=18)}
  begin
    dlam := i;
    som_series(fb, fa2, fa4, fc1, fc3, dlam);
    suma2 := suma2 + 2.0 * fa2;
    suma4 := suma4 + 2.0 * fa4;
    sumb := sumb + 2.0 * fb;
    sumc1 := sumc1 + 2.0 * fc1;
    sumc3 := sumc3 + 2.0 * fc3;
  end;

  dlam := 90.0;
  som_series(fb, fa2, fa4, fc1, fc3, dlam);
  suma2 := suma2 + fa2;
  suma4 := suma4 + fa4;
  sumb := sumb + fb;
  sumc1 := sumc1 + fc1;
  sumc3 := sumc3 + fc3;
  a2 := suma2 / 30.0;
  a4 := suma4 / 60.0;
  b := sumb / 30.0;
  c1 := sumc1 / 15.0;
  c3 := sumc3 / 45.0;
  result := (OK);
end;


function
  somfor(lon, lat: double; var y, x: double): Integer;
{long
somfor(double lon, double lat, double *y, double *x)

             /* (I) Longitude 		*/
             /* (I) Latitude 		*/
            /* (O) X projection coordinate 	*/
            /* (O) Y projection coordinate 	*/}
var
  n, i, l: Integer;
  delta_lon,
    rlm, tabs, tlam, xlam, c, xlamt, ab2, sc1, ab1, xlamp, sav,
    d, sdsq, sd, tanlg, xtan, tphi, dp, dd, ds, rlm2,
    scl, tlamp, conv, delta_lat, radlt, radln,
    temp: Double;
{char   errorbuf[80];}
label L230, L240, L250, L260, L300;
begin
//* Forward equations  -----------------*/
  conv := 1.E-7;
  delta_lat := lat;
  delta_lon := lon - lon_center;

//* Test for latitude and longitude approaching 90 degrees

  if (delta_lat > 1.570796) then delta_lat := 1.570796;
  if (delta_lat < -1.570796) then delta_lat := -1.570796;
  radlt := delta_lat;
  radln := delta_lon;
  if (delta_lat >= 0.0) then tlamp := PI / 2.0
  else tlamp := 1.5 * PI;
  if (start <> 0.0) then tlamp := 2.5 * PI;
  n := 0;

  L230: sav := tlamp;
  l := 0;
  xlamp := radln + p21 * tlamp;
  ab1 := cos(xlamp);
  if (abs(ab1) < conv) then xlamp := xlamp - 1.E-7;
  if xlamp > 0 then {shut compiler up};
  if (ab1 >= 0.0) then scl := 1.0 else scl := -1.0;
  ab2 := tlamp - (scl) * sin(tlamp) * HALF_PI;
  L240: xlamt := radln + p21 * sav;
  c := cos(xlamt);
  if (abs(c) < 1.E-7) then xlamt := xlamt - 1.E-7;
  xlam := (((1.0 - es) * tan(radlt) * sa) + sin(xlamt) * ca) / c;
  tlam := arctan(xlam);
  tlam := tlam + ab2;
  tabs := abs(sav) - abs(tlam);
  if (abs(tabs) < conv) then goto L250;
  l := l + 1;
  if (l > 50) then goto L260;
  sav := tlam;
  goto L240;

//* Adjust for confusion at beginning and end of landsat orbits

  L250: rlm := PI * LANDSAT_RATIO;
  rlm2 := rlm + 2.0 * PI;
       {n++;} inc(n);
  if (n >= 3) then goto L300;
  if ((tlam > rlm) and (tlam < rlm2)) then goto L300;
  if (tlam < rlm) then tlamp := 2.50 * PI;
  if (tlam >= rlm2) then tlamp := HALF_PI;
  goto L230;
  L260: {sprintf(errorbuf,'50 iterations without conv\n');}
  {PError('errorbuf', 'som-forward'); }
  DoProjError(214,214,0,0,0);
  result := (214);
  exit;
//* tlam computed - now compute tphi

  L300: ds := sin(tlam);
  dd := ds * ds;
  dp := sin(radlt);
  tphi := arcsin(((1.0 - es) * ca * dp - sa * cos(radlt) * sin(xlamt))
    / sqrt(1.0 - es * dp * dp));

//* compute x and y

  xtan := (PI / 4.0) + (tphi / 2.0);
  tanlg := ln(tan(xtan));
  sd := sin(tlam);
  sdsq := sd * sd;
  s := p21 * sa * cos(tlam) * sqrt((1.0 + t * sdsq) / ((1.0 + w *
    sdsq) * (1.0 + q * sdsq)));
  d := sqrt(xj * xj + s * s);
  x := b * tlam + a2 * sin(2.0 * tlam) + a4 * sin(4.0 * tlam) - tanlg
    * s / d;
  x := a * x;
  y := c1 * sd + c3 * sin(3.0 * tlam) + tanlg * xj / d;
  y := a * y;

//* Negate x & swap x,y
  temp := x;
  x := y + false_easting;
  y := temp + false_northing; ;
  result := (OK);
end;





function
  sominvint(r_major, r_minor: double;
  satnum, path: Integer;
  alf_in, lon, false_east, false_north,
  time: double; start1, flag: Integer): Integer;
{long
sominvint(double r_major, double r_minor, long int satnum, long int path, double alf_in, double lon, double false_east, double false_north, double time, long int start1, long int flag)
                 /* major axis				*/
                 /* minor axis				*/
               /* Landsat satellite number (1,2,3,4,5) */
             /* Landsat path number */
                    /* x offset in meters			*/
                     /* y offset in meters}
var
  i: Integer;
  alf, e2c, e2s, one_es,
    dlam, fb, fa2, fa4, fc1, fc3, suma2, suma4, sumc1, sumc3, sumb:
      Double;
begin
//* Place parameters in static storage for common use

  false_easting := false_east;
  false_northing := false_north;
  a := r_major;
  b := r_minor;
  es := 1.0 - SQR(r_minor / r_major);

  if (flag <> 0) then
  begin
    alf := alf_in;
    lon_center := lon;
    p21 := time / 1440.0;
    start := start1;
  end
  else
  begin
    if (satnum < 4) then
    begin
      alf := 99.092 * D2R;
      p21 := 103.2669323 / 1440.0;
      lon_center := (128.87 - (360.0 / 251.0 * path)) * D2R;
    end
    else
    begin
      alf := 98.2 * D2R;
      p21 := 98.8841202 / 1440.0;
      lon_center := (129.30 - (360.0 / 233.0 * path)) * D2R;
    {lon_center = (129.30557714 - (360.0/233.0 * path)) * D2R;}
    end;
    start := 0.0;
  end;

//* Report parameters to the user (to device set up prior to this call)

  ptitle('SPACE OBLIQUE MERCATOR');
  radius2(a, b);
  GenrptLong(path, 'Path Number:    ');
  GenrptLong(satnum, 'Satellite Number:    ');
  genrpt(alf * R2D, 'Inclination of Orbit:    ');
  genrpt(lon_center * R2D, 'Longitude of Ascending Orbit:    ');
  offsetp(false_easting, false_northing);
  genrpt(LANDSAT_RATIO, 'Landsat Ratio:    ');

  ca := cos(alf);
  if (abs(ca) < 1.E-9) then ca := 1.E-9;
  sa := sin(alf);
  e2c := es * ca * ca;
  e2s := es * sa * sa;
  w := (1.0 - e2c) / (1.0 - es);
  w := w * w - 1.0;
  one_es := 1.0 - es;
  q := e2s / one_es;
  t := (e2s * (2.0 - es)) / (one_es * one_es);
  u := e2c / one_es;
  xj := one_es * one_es * one_es;
  dlam := 0.0;
  som_series(fb, fa2, fa4, fc1, fc3, dlam);
  suma2 := fa2;
  suma4 := fa4;
  sumb := fb;
  sumc1 := fc1;
  sumc3 := fc3;
{for(i=9;i<=81;i+=18)}
  i := 9; repeat
    begin
      dlam := i;
      som_series(fb, fa2, fa4, fc1, fc3, dlam);
      suma2 := suma2 + 4.0 * fa2;
      suma4 := suma4 + 4.0 * fa4;
      sumb := sumb + 4.0 * fb;
      sumc1 := sumc1 + 4.0 * fc1;
      sumc3 := sumc3 + 4.0 * fc3;
      inc(i, 18);
  end until i <= 81;

{for(i=18; i<=72; i+=18)}
  i := 18; repeat
    begin
      dlam := i;
      som_series(fb, fa2, fa4, fc1, fc3, dlam);
      suma2 := suma2 + 2.0 * fa2;
      suma4 := suma4 + 2.0 * fa4;
      sumb := sumb + 2.0 * fb;
      sumc1 := sumc1 + 2.0 * fc1;
      sumc3 := sumc3 + 2.0 * fc3;
      inc(i, 18);
  end until i <= 72;

  dlam := 90.0;
  som_series(fb, fa2, fa4, fc1, fc3, dlam);
  suma2 := suma2 + fa2;
  suma4 := suma4 + fa4;
  sumb := sumb + fb;
  sumc1 := sumc1 + fc1;
  sumc3 := sumc3 + fc3;
  a2 := suma2 / 30.0;
  a4 := suma4 / 60.0;
  b := sumb / 30.0;
  c1 := sumc1 / 15.0;
  c3 := sumc3 / 45.0;
  result := (OK);
end;



function
  sominv(y, x: double; var lon, lat: double): Integer;
{long sominv(double y, double x, double *lon, double *lat)
                        /* (I) X projection coordinate */
                        /* (I) Y projection coordinate */
                        /* (O) Longitude */
                        /* (O) Latitude */}
var
  tlon, conv, sav, sd, sdsq, blon, dif, st, defac, actan, tlat, dd,
    bigk, bigk2, xlamt,
    sl, scl, dlat, dlon, temp: double;
  inumb: Integer;
begin
//* Inverse equations. Begin inverse computation with approximation for tlon.
//   Solve for transformed long.
//  ---------------------------*/
  temp := y; y := x - false_easting; x := temp - false_northing;
  tlon := x / (a * b);
  conv := 1.E-9;
  for inumb := 0 to 50 do
  begin
    sav := tlon;
    sd := sin(tlon);
    sdsq := sd * sd;
    s := p21 * sa * cos(tlon) * sqrt((1.0 + t * sdsq) / ((1.0 + w *
      sdsq) * (1.0 + q * sdsq)));
    blon := (x / a) + (y / a) * s / xj - a2 * sin(2.0 * tlon) - a4 *
      sin(4.0 * tlon) - (s / xj) * (c1 *
      sin(tlon) + c3 * sin(3.0 * tlon));
    tlon := blon / b;
    dif := tlon - sav;
    if (abs(dif) < conv) then break;
    if (inumb >= 50) then
    begin
      {PError('50 iterations without convergence', 'som-inverse');}
      DoProjError(215,214,0,0,0);
      result := (214);
      exit;
    end;
  end;


//* Compute transformed lat.

  st := sin(tlon);
  defac := exp(sqrt(1.0 + s * s / xj / xj) * (y / a - c1 * st - c3 *
    sin(3.0 * tlon)));
  actan := arctan(defac);
  tlat := 2.0 * (actan - (PI / 4.0));

//* Compute geodetic longitude

  dd := st * st;
  if (abs(cos(tlon)) < 1.E-7) then tlon := tlon - 1.E-7;
  bigk := sin(tlat);
  bigk2 := bigk * bigk;
  xlamt := arctan(((1.0 - bigk2 / (1.0 - es)) * tan(tlon) * ca - bigk
    * sa * sqrt((1.0 + q * dd)
    * (1.0 - bigk2) - bigk2 * u) / cos(tlon)) / (1.0 - bigk2 * (1.0 +
      u)));

//* Correct inverse quadrant

  if (xlamt >= 0.0) then sl := 1.0 else sl := -1.0;
  if (cos(tlon) >= 0.0) then scl := 1.0 else scl := -1.0;
  xlamt := xlamt - ((PI / 2.0) * (1.0 - scl) * sl);
  dlon := xlamt - p21 * tlon;

//* Compute geodetic latitude

  if (abs(sa) < 1.E-7) then
    dlat := arcsin(bigk / sqrt((1.0 - es) * (1.0 - es) + es * bigk2))
  else
    dlat := arctan((tan(tlon) * cos(xlamt) - ca * sin(xlamt)) / ((1.0 -
    es) * sa));
  lon := AdjustLon(dlon + lon_center);
  lat := dlat;
  result := (OK);
end;



procedure som_series(var fb, fa2, fa4, fc1, fc3, dlam: double);
{Series to calculate a,b,c coefficients to convert from transform
   latitude,longitude to Space Oblique Mercator (SOM) rectangular coordinates
   Mathematical analysis by John Snyder 6/82
  --------------------------------------------------------------------------*/
static void som_series(double *fb, double *fa2, double *fa4,
double *fc1, double *fc3, double *dlam)}
var sd, sdsq, h, sq, fc: double;
begin
  dlam := dlam * 0.0174532925; //* Convert dlam to radians */
  sd := sin(dlam);
  sdsq := sd * sd;
  s := p21 * sa * cos(dlam) * sqrt((1.0 + t * sdsq) / ((1.0 + w *
    sdsq) * (1.0 + q * sdsq)));
  h := sqrt((1.0 + q * sdsq) / (1.0 + w * sdsq)) * (((1.0 + w * sdsq)
    / ((1.0 + q * sdsq) * (1.0 +
    q * sdsq))) - p21 * ca);
  sq := sqrt(xj * xj + s * s);
  fb := (h * xj - s * s) / sq;
  fa2 := fb * cos(2.0 * dlam);
  fa4 := fb * cos(4.0 * dlam);
  fc := s * (h + xj) / sq;
  fc1 := fc * cos(dlam);
  fc3 := fc * cos(3.0 * dlam);
end;


end.
