{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzWiv;

{*******************************************************************
NAME                            WAGNER IV

PURPOSE:	Transforms input longitude and latitude to Easting and
  Northing for the Wagner IV projection.  The
  longitude and latitude must be in radians.  The Easting
  and Northing values will be returned in meters.

PROGRAMMER              DATE
----------              ----
D. Steinwand, EROS      May, 1991

This function was implemented with formulas supplied by John P. Snyder.

ALGORITHM REFERENCES

1.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.

2.  Snyder, John P., Personal correspondence, January 1991.
*******************************************************************************/
#include "gctpc/cproj.h"
#include "gctpc/report.h"}
interface

uses Math;
//Variables common to all subroutines in this code file
var
  lon_center, //* Center longitude (projection center) */
    R, //* Radius of the earth (sphere) */
    false_easting, //* x offset				*/
    false_northing: Double; //* y offset				*/


function
  wivforint(r, center_long, false_east, false_north: double):
    Integer;

function
  wivfor(lon, lat: double; var x, y: double): Integer;

function
  wivinvint(r, center_long, false_east, false_north: double):
    Integer;

function
  wivinv(x, y: double; var lon, lat: double): Integer;

implementation
uses buzGlobal;


function
  wivforint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the Wagner IV projection
long
wivforint(double r, double center_long, double false_east, double false_north)
             /* (I) Radius of the earth (sphere) */
                     /* (I) Center longitude */
                    /* x offset
                     /* y offset}
begin
//* Place parameters in static storage for common use
  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;
//* Report parameters to the user
  ptitle('WAGNER IV');
  radius(r);
  cenlon(center_long);
  offsetp(false_east, false_north);
  result := (OK);
end;


function
  wivfor(lon, lat: double; var x, y: double): Integer;
{Wagner IV forward equations--mapping lat,long to x,y
long
wivfor(double lon, double lat, double *x, double *y)
              /* (I) Longitude */
              /* (I) Latitude */
             /* (O) X projection coordinate */
             /* (O) Y projection coordinate */}
var
{double AdjustLon(double x);
/* Function to adjust longitude to -180 - 180 */   }
  delta_lon, //* Delta longitude (Given longitude - center */
    theta,
    delta_theta,
    con: double;
  i: Integer;
begin
//* Forward equations

  delta_lon := AdjustLon(lon - lon_center);
  theta := lat;
  con := 2.9604205062 * sin(lat);

//* Iterate using the Newton-Raphson method to find theta

  for i := 0 to 30 do
  begin
    delta_theta := -(theta + sin(theta) - con) / (1.0 + cos(theta));
    theta := theta + delta_theta;
    if (abs(delta_theta) < EPSLN) then break;
   { if (i >= 30) perror("Iteration failed to converge","wagneriv-forward");}
    if (i >= 30) then
    begin
     { PError('Iteration failed to converge', 'Wagneriv-Forward'); }
      DoProjError(243,241,0,0,0);
      result := (OK);
      exit;
    end;
  end;
  theta := theta / 2.0;
  x := 0.86310 * R * delta_lon * cos(theta) + false_easting;
  y := 1.56548 * R * sin(theta) + false_northing;
  result := (OK);
end;



function
  wivinvint(r, center_long, false_east, false_north: double):
    Integer;
{Initialize the Wagner IV projection
long wivinvint(double r, double center_long,
double false_east, double false_north)
             /* (I) Radius of the earth (sphere) */
                     /* (I) Center longitude */
                    /* x offset
                     /* y offset}
begin
//* Place parameters in static storage for common use

  R := r;
  lon_center := center_long;
  false_easting := false_east;
  false_northing := false_north;

//* Report parameters to the user

  ptitle('WAGNER IV');
  radius(r);
  cenlon(center_long);
  offsetp(false_east, false_north);
  result := (OK);
end;


function
  wivinv(x, y: double; var lon, lat: double): Integer;
{Wagner IV inverse equations--mapping x,y to lat,long
long
wivinv(double x, double y, double *lon, double *lat)
           /* (I) X projection coordinate */
           /* (I) Y projection coordinate */
              /* (O) Longitude */
              /* (O) Latitude */}
var theta: double;
begin
{double AdjustLon(double x);
 /* Function to adjust longitude to -180 - 180 */}
//* Inverse equations

  x := x - false_easting;
  y := y - false_northing;
  theta := arcsin(y / (1.56548 * R));
  lon := AdjustLon(lon_center + (x / (0.86310 * R * cos(theta))));
  lat := arcsin((2.0 * theta + sin(2.0 * theta)) / 2.9604205062);
  result := (OK);
end;



end.
